clc; clear;

% Thin plate spline smoothing of local volatility
% By Fabrice Douglas Rouah, FRouah.com and Volopta.com

BSC = @(s,K,rf,q,v,T) (s*exp(-q*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T)) - K*exp(-rf*T)*normcdf((log(s/K) + (rf-q+v^2/2)*T)/v/sqrt(T) - v*sqrt(T)));

% Input the SPY implied volatilities from calls
MktIV = [ ...
   27.8000   26.3800   25.3200   25.1800
   24.7700   24.0200   23.6400   23.6900
   21.8600   21.5800   22.0300   22.3900
   18.7800   19.3000   20.4700   20.9800
   15.7200   17.1200   18.9400   19.7000
   13.3400   15.1700   17.4800   18.4900
   13.2300   13.7300   16.1800   17.3600]./100;
[NK NT] = size(MktIV);

% Maturities and strikes
T = [45 98 261 348]./365;
K = [120   125   130   135   140   145   150];

% Spot, risk free rate, dividend yield
S0 = 137.14;
r = 0.0010;
q  = 0.0068;
PutCall = 'C';

% Market prices of calls
for k=1:NK
    for t=1:NT
        MktPrice(k,t) = BSC(S0,K(k),r,q,MktIV(k,t),T(t));
    end
end

%% Input the Heston prices based on the following parameters
% kappa =  1.6651;
% theta =  0.0955;
% sigma =  0.9699;
% v0    =  0.0344;
% rho   = -0.7758;
% lambda = 0;

HPrice = [...
   17.6083   18.6836   21.6618   23.0857;
   13.0038   14.3362   17.7163   19.2885;
    8.6895   10.2663   14.0093   15.7149;
    4.8766    6.6013   10.6076   12.4142;
    1.9254    3.5504    7.5935    9.4432;
    0.4255    1.4680    5.0657    6.8647;
    0.0638    0.5007    3.1198    4.7390];

%% Find the implied volatilities
% Settings for bisection algorithm
a = 0.01;
b = 2.0;
Tol = 1e-5;
MaxIter = 1e5;

for t=1:NT
    for k=1:NK
        MktIV(k,t) = BisecBSV(PutCall,S0,K(k),r,q,T(t),a,b,MktPrice(k,t),Tol,MaxIter);
        IV(k,t) = BisecBSV('C',S0,K(k),r,q,T(t),a,b,HPrice(k,t),Tol,MaxIter);
    end
end

%% Obtain the local volatility from the implied volatility
dK = 0.1;
dT = 0.001;
LV = getLVfromIV(S0,K,T,IV,dK,dT,r,q);


%% Smooth the local volatility using thin plate splines
% Build the mesh where the smoothed LV will be constructed
Nxi = 50;
Nyi = 25;
dK = (K(end)-K(1))/(Nxi-1);
dT = (T(end)-T(1))/(Nyi-1);
Ki = K(1):dK:K(end);
Ti = T(1):dT:T(end);

% Smoothing parameter
lambda = 0.1;

% Request TPS smoothed Local vol and detailed output
detail = 'Y';
LVi = TPS(K,T,LV,Ki,Ti,lambda,detail);

%% Plot everything
surf(Ki,Ti,LVi')
hold on
mesh(K,T,IV','Marker','o','MarkerFaceColor','k')
alpha(0.25)
hold off
xlabel('Strike')
ylabel('Maturity')
title('Implied vol (dots) and Local vol (Surface)')
view([1 -1 1])

        